/*****************************************************************************
*				Copyright (c) Gemplus Card International 1997
*
* Name		: reader.c
*
* Description : SmartCard reader interface

  Author	: JL GIRAUD
  Compiler	: Visual C++, GNU gcc
  Host		: Windows 9x, NT, Linux
*
* $Id: reader.c,v 1.11 2000/12/11 17:53:29 rousseau Exp $
*
******************************************************************************
*
* Warning	:
*
* Remarks	:
*
*****************************************************************************/

#ifndef WIN32
#include <fcntl.h>
#include <termios.h>
#include <unistd.h>
#include <sys/file.h>
#endif

#include <stdio.h>
#include <errno.h>
#include <string.h>

#include "reader.h"
#include "tlpreader.h"
#include "rftlpreader.h"
#include "c2bin.h"

int rerrno;

char *rd_errors[] =
{
	"no error",		/* RD_NO_ERROR */
	"unknown command",
	"driver limitation",
	"wrong arg length",
	"protocol error",
	"wrong ATR",
	"wrong rd ins",
	"message too long",
	"read error",
	"card off",
	"prog volt error",
	"unknown protocol",
	"length error",
	"wrong number of parameters",
	"unknown card protocol",
	"card mute",
	"card parity error",
	"card wrong procedure byte",
	"transfert to card interrupted",
	"card removed",
	"short circuit",
	"card absent",
	"read on serial port failed",
	"reader complained about a transmission error",
	"write to serial port failed",
	"response too long",
	"reader mute",
	"status too long",
	"useless error",
	"chaining error",
	"response truncated",
	"unknown error"			/* RD_UNKNOWN_ERROR */
};

char iso7816_error[ISO7816_ERROR_TEXT_SIZE];

#ifdef _DEBUG
static void reader_send_error(char* info_string)
{
	fprintf(stderr,"%s: %s\n", info_string, strerror(errno));
} /* reader_send_error */
#endif

/*****************************************************************************/

/* internal functions */

int init_serial(ReaderDesc *ad_rd, char *device);
void set_reader_type(ReaderDesc *ad_rd, rd_type reader_type);

/*****************************************************************************/


t_rd_error GetRerrno(void)
{
	return rerrno;
} /* GetRerrno */

/*****************************************************************************/


char *GetRerrStr(t_rd_error nErr)
{
	if (nErr >= 0 && nErr < RD_NUM_ERRORS)
		return rd_errors[nErr];
	else
		return rd_errors[RD_UNKNOWN_ERROR];
} /* GetRerrStr */

/*****************************************************************************/


int InitReader(ReaderDesc *ad_rd, rd_type type, char *device)
{
	int ret;

	switch (type)
	{
		case TLP:
		case RFTLP:
			ret = init_serial(ad_rd, device);
			if (ret)
				return ret;

			set_reader_type(ad_rd, type);
			break;

		case UNKNOWN:
		default:
			/* should no occur */
			memset(ad_rd, 0, sizeof(*ad_rd));
			rerrno = 
			ret = -1;
	}

	return ret;
} /* InitReader */

/*****************************************************************************/


int init_serial(ReaderDesc *ad_rd, char *device)
{
#ifndef WIN32	//////////////////////  Linux  /////////////////////////
	struct termios newtio;

	ad_rd->fd = open(device, O_RDWR | O_NOCTTY );
	if (ad_rd->fd <0)
	{
#ifdef _DEBUG
		reader_send_error(device);
#endif
		return -1;
	}

	/* test if the device is not used by an other process*/
	if (lockf (ad_rd->fd,F_TLOCK,0) == -1)
		return -1;

	tcgetattr(ad_rd->fd,&(ad_rd->oldtio)); /* save current port settings */

	bzero(&newtio, sizeof(newtio));
	/* BAUDRATE = 9600							*/
	/* CS8	= 8n1 (8bit,no parity,1 stopbit*/
	/* CLOCAL= local connection, no modem control*/
	/* CREAD	= enable receiving characters*/
	newtio.c_cflag = BAUDRATE |  CS8 | CLOCAL | CREAD;
	newtio.c_iflag = IGNPAR;
	newtio.c_oflag = 0;

	/* set input mode (non-canonical, no echo,...) */
	newtio.c_lflag = 0;

	newtio.c_cc[VTIME]    = 0;   /* inter-character timer unused */
	newtio.c_cc[VMIN]     = 1;   /* blocking read until 1 chars received */

	tcflush(ad_rd->fd, TCIFLUSH);
	if (tcsetattr(ad_rd->fd,TCSANOW,&newtio))
	{
#ifdef _DEBUG
		reader_send_error("Can't set attributes");
#endif
		close(ad_rd->fd);
		return -1;
	}

#else	//////////////////////  WIN32  /////////////////////////
	DCB dcb;
	BOOL fSuccess;
	const int TIME_OUT = 2000;
	COMMTIMEOUTS stTime;

	ad_rd->fd = CreateFile
					(device,GENERIC_READ | GENERIC_WRITE,
					0,	// comm devices must be opened w/exclusive-access
					NULL,	// no security attributes
					OPEN_EXISTING,// comm devices must use OPEN_EXISTING
					0,	// not overlapped I/O
					NULL	// hTemplate must be NULL for comm devices
					);
	
	if (ad_rd->fd  == INVALID_HANDLE_VALUE)
	{ // Handle the error.
#ifdef _DEBUG
		reader_send_error(device);
#endif
		return -1;
	}

	stTime.ReadIntervalTimeout = TIME_OUT;
	stTime.ReadTotalTimeoutConstant = TIME_OUT;
	stTime.ReadTotalTimeoutMultiplier = TIME_OUT;
	stTime.WriteTotalTimeoutConstant = TIME_OUT;
	stTime.WriteTotalTimeoutMultiplier = TIME_OUT;

	SetCommTimeouts (ad_rd->fd, &stTime);

	/* save current port settings */
	fSuccess = GetCommState(ad_rd->fd, &ad_rd->oldtio);
	if (!fSuccess)
	{
	// Handle the error.
#ifdef _DEBUG
		reader_send_error(device);
#endif
		return -1;
	}

	// Configuring the device
	dcb = ad_rd->oldtio ;
	dcb.BaudRate = BAUDRATE;
	dcb.ByteSize = 8;
	dcb.Parity = NOPARITY;  
	dcb.StopBits = ONESTOPBIT;

	fSuccess = SetCommState(ad_rd->fd , &dcb);
	if (!fSuccess)
	{
#ifdef _DEBUG
		reader_send_error(device);
#endif
		CloseHandle(ad_rd->fd);
		return -1;
	}

	PurgeComm (ad_rd->fd,
		PURGE_TXABORT|PURGE_RXABORT|PURGE_TXCLEAR|PURGE_RXCLEAR);

#endif // WIN32

	return 0;
} /* init_serial */

/*****************************************************************************/


void CloseReader(ReaderDesc *ad_rd)
{
#ifndef WIN32
	tcsetattr(ad_rd->fd,TCSANOW,&(ad_rd->oldtio));
	lockf (ad_rd->fd,F_ULOCK,0);
	close(ad_rd->fd);
#else
	SetCommState(ad_rd->fd , &ad_rd->oldtio );
	CloseHandle(ad_rd->fd);
#endif
} /* CloseReader */

/*****************************************************************************/


void set_reader_type(ReaderDesc *ad_rd, rd_type reader_type)
{
	switch (reader_type)
	{
		case RFTLP:
			ad_rd->TrError = RFTLPTrError;
			ad_rd->ResetCardBinary = RFTLPResetCardBinary;
			ad_rd->ResetCardAscii = RFTLPResetCardAscii;
			ad_rd->PowerDown = RFTLPPowerDown;
			ad_rd->SendIncomingAPDUBinary = RFTLPSendIncomingAPDUBinary;
			ad_rd->SendOutgoingAPDUBinary = RFTLPSendOutgoingAPDUBinary;
			ad_rd->SendIncomingOutgoingAPDUBinary =
				RFTLPSendIncomingOutgoingAPDUBinary;	
			ad_rd->SendIncomingAPDUAscii  = RFTLPSendIncomingAPDUAscii;
			ad_rd->SendOutgoingAPDUAscii  = RFTLPSendOutgoingAPDUAscii;
			ad_rd->SendIncomingOutgoingAPDUAscii =
				RFTLPSendIncomingOutgoingAPDUAscii;	
			break;

		case TLP:
			ad_rd->TrError = TLPTrError;
			ad_rd->ResetCardBinary = TLPResetCardBinary;
			ad_rd->ResetCardAscii = TLPResetCardAscii;
			ad_rd->PowerDown = TLPPowerDown;
			ad_rd->SendIncomingAPDUBinary = TLPSendIncomingAPDUBinary;
			ad_rd->SendOutgoingAPDUBinary = TLPSendOutgoingAPDUBinary;
			ad_rd->SendIncomingOutgoingAPDUBinary =
				TLPSendIncomingOutgoingAPDUBinary;
			ad_rd->SendIncomingAPDUAscii  = TLPSendIncomingAPDUAscii;
			ad_rd->SendOutgoingAPDUAscii  = TLPSendOutgoingAPDUAscii;
			ad_rd->SendIncomingOutgoingAPDUAscii =
				TLPSendIncomingOutgoingAPDUAscii;
			break;

		case UNKNOWN:
		default:
			ad_rd->TrError = NULL;
			ad_rd->ResetCardBinary = NULL;
			ad_rd->ResetCardAscii = NULL;
			ad_rd->PowerDown = NULL;
			ad_rd->SendIncomingAPDUBinary = NULL;
			ad_rd->SendOutgoingAPDUBinary = NULL;
			ad_rd->SendIncomingOutgoingAPDUBinary = NULL;
			ad_rd->SendIncomingAPDUAscii  = NULL;
			ad_rd->SendOutgoingAPDUAscii  = NULL;
			ad_rd->SendIncomingOutgoingAPDUAscii = NULL;
			break;
	}
} /* set_reader_type */

/*****************************************************************************/


int PowerDown(ReaderDesc *ad_rd)
{
	return ad_rd->PowerDown(ad_rd->fd);
} /* PowerDown */

/*****************************************************************************/


int ResetCardBinary(ReaderDesc *ad_rd, unsigned char* atr, int length)
{
	return ad_rd->ResetCardBinary(ad_rd->fd, atr, length);
} /* ResetCardBinary */

/*****************************************************************************/


int ResetCardAscii(ReaderDesc *ad_rd, char* atr, int length)
{
	return ad_rd->ResetCardAscii(ad_rd->fd, atr, length);
} /* ResetCardAscii */

/*****************************************************************************/


int SendIncomingAPDUAscii(ReaderDesc *ad_rd, const char* APDU_Data,
			char* ad_response, int resp_length)
{
	return ad_rd->SendIncomingAPDUAscii(ad_rd->fd,APDU_Data, ad_response,
					resp_length);
} /* SendIncomingAPDUAscii */

/*****************************************************************************/


int SendOutgoingAPDUAscii(ReaderDesc *ad_rd, const char* APDU_Data,
			char* ad_response, int resp_length)
{
	return ad_rd->SendOutgoingAPDUAscii(ad_rd->fd, APDU_Data,ad_response,
					resp_length);
} /* SendOutgoingAPDUAscii */


/*****************************************************************************/


int SendIncomingOutgoingAPDUAscii(ReaderDesc *ad_rd, const char* APDU_Data,
				char* ad_response, int resp_length)
{
	return ad_rd->SendIncomingOutgoingAPDUAscii(ad_rd->fd, APDU_Data,
						ad_response, resp_length);
} /* SendIncomingOutgoingAPDUAscii */

/*****************************************************************************/


int SendOutgoingAPDUBinary(ReaderDesc *ad_rd,
			unsigned char CLA,
			unsigned char INS,
			unsigned char P1,
			unsigned char P2,
			unsigned char L,
			unsigned char *data_out,
			unsigned char *ad_SW1,
			unsigned char *ad_SW2)
{
	return ad_rd-> SendOutgoingAPDUBinary(ad_rd->fd, CLA, INS, P1, P2, L,
					data_out, ad_SW1, ad_SW2);
} /* SendOutgoingAPDUBinary */

/*****************************************************************************/


int SendIncomingAPDUBinary(ReaderDesc *ad_rd,
			unsigned char CLA,
			unsigned char INS,
			unsigned char P1,
			unsigned char P2,
			unsigned char L,
			const unsigned char *data_in,
			unsigned char *ad_SW1,
			unsigned char *ad_SW2)
{
	return ad_rd-> SendIncomingAPDUBinary(ad_rd->fd, CLA, INS, P1, P2, L,
					data_in, ad_SW1, ad_SW2);
} /* SendIncomingAPDUBinary */

/*****************************************************************************/


int SendIncomingOutgoingAPDUBinary(ReaderDesc *ad_rd,
				unsigned char CLA,
				unsigned char INS,
				unsigned char P1,
				unsigned char P2,
				unsigned char L_in,
				unsigned char *data_in,
				unsigned char L_out,
				unsigned char *data_out,
				unsigned char *ad_SW1,
				unsigned char *ad_SW2)
{
	return ad_rd->SendIncomingOutgoingAPDUBinary(ad_rd->fd, CLA, INS, P1, P2,
						L_in, data_in,
						L_out, data_out,
						ad_SW1, ad_SW2);
} /* SendIncomingOutgoingAPDUBinary */

/*****************************************************************************/

char *ISO7816_ErrorStr(unsigned char sw1, unsigned char sw2)
{
	char *ptr = iso7816_error;

	/* status words */
#ifdef WIN32
	_snprintf(ptr, ISO7816_ERROR_TEXT_SIZE-1, "%02X%02X: ", sw1, sw2);
#else
	snprintf(ptr, ISO7816_ERROR_TEXT_SIZE-1, "%02X%02X: ", sw1, sw2);
#endif

	/* go to the end of the string */
	ptr += strlen(ptr);
	
	if (sw1 == 0x90 && sw2 == 0x00)
		strncat(ptr, "Normal processing",
			ISO7816_ERROR_TEXT_SIZE-strlen(iso7816_error)-1);

	if (sw1 == 0x61)
#ifdef WIN32
		_snprintf(ptr, ISO7816_ERROR_TEXT_SIZE-strlen(iso7816_error)-1,
			"%d bytes of response still available.", sw2);
#else
		snprintf(ptr, ISO7816_ERROR_TEXT_SIZE-strlen(iso7816_error)-1,
			"%d bytes of response still available.", sw2);
#endif

	if (sw1 == 0x62)
	{
		strncat(ptr, "State of non-volatile memory unchanged. ", 
			ISO7816_ERROR_TEXT_SIZE-strlen(iso7816_error)-1);

		if (sw2 == 0x00)
			strncat(ptr, "No informatin given.",
				ISO7816_ERROR_TEXT_SIZE-1-strlen(iso7816_error));

		if (sw2 == 0x81)
			strncat(ptr, "Part of returned data my be corrupted.",
				ISO7816_ERROR_TEXT_SIZE-1-strlen(iso7816_error));

		if (sw2 == 0x82)
			strncat(ptr,
				"End of file/record reached before reading Le bytes.",
				ISO7816_ERROR_TEXT_SIZE-1-strlen(iso7816_error));

		if (sw2 == 0x83)
			strncat(ptr, "Selected file invalidated.",
				ISO7816_ERROR_TEXT_SIZE-1-strlen(iso7816_error));

		if (sw2 == 0x84)
			strncat(ptr, "FCI not formated according to 5.1.5.",
				ISO7816_ERROR_TEXT_SIZE-1-strlen(iso7816_error));
	}

	if (sw1 == 0x63)
	{
		strncat(ptr, "State of non-volatile memory changed. ",
			ISO7816_ERROR_TEXT_SIZE-strlen(iso7816_error)-1);

		/* increment pointer for the snprintf() in the last case */
		ptr += strlen(ptr);

		if (sw2 == 0x00)
			strncat(ptr, "No informatin given.",
				ISO7816_ERROR_TEXT_SIZE-1-strlen(iso7816_error));

		if (sw2 == 0x81)
			strncat(ptr, "File filled up by the last write.",
				ISO7816_ERROR_TEXT_SIZE-1-strlen(iso7816_error));

		if ((sw2 & 0xF0) == 0xC0)
#ifdef WIN32
			_snprintf(ptr, 
				ISO7816_ERROR_TEXT_SIZE-strlen(iso7816_error)-1,
				"Counter %d.", sw2 & 0x0F);
#else
			snprintf(ptr, 
				ISO7816_ERROR_TEXT_SIZE-strlen(iso7816_error)-1,
				"Counter %d.", sw2 & 0x0F);
#endif
	}

	if (sw1 == 0x64 && sw2 == 0x00)
		strncat(ptr,
			"State of non-volatile memory unchanged.",
			ISO7816_ERROR_TEXT_SIZE-strlen(iso7816_error)-1);

	if (sw1 == 0x65)
	{
		strncat(ptr,
			"State of non-volatile memory changed. ",
			ISO7816_ERROR_TEXT_SIZE-strlen(iso7816_error)-1);

		if (sw2 == 0x81)
			strncat(ptr, "Memory failure.",
				ISO7816_ERROR_TEXT_SIZE-1-strlen(iso7816_error));
	}

	if (sw1 == 0x66)
		strncat(ptr,
			"Reserved for security-related issues.",
			ISO7816_ERROR_TEXT_SIZE-strlen(iso7816_error)-1);

	if (sw1 == 0x67 && sw2 == 0x00)
		strncat(ptr, "Wrong length.",
			ISO7816_ERROR_TEXT_SIZE-strlen(iso7816_error)-1);

	if (sw1 == 0x68 && sw2 == 0x00)
	{
		strncat(ptr, "Functions in CLA not supported. ",
			ISO7816_ERROR_TEXT_SIZE-strlen(iso7816_error)-1);

		if (sw2 == 0x81)
			strncat(ptr, "Logical channel not supported.",
				ISO7816_ERROR_TEXT_SIZE-1-strlen(iso7816_error));

		if (sw2 == 0x82)
			strncat(ptr, "Secure messaging not supported.",
				ISO7816_ERROR_TEXT_SIZE-1-strlen(iso7816_error));
	}

	if (sw1 == 0x69)
	{
		strncat(ptr, "Command not allowed. ",
			ISO7816_ERROR_TEXT_SIZE-strlen(iso7816_error)-1);

		if (sw2 == 0x81)
			strncat(ptr, "Command incompatible with file structure.",
				ISO7816_ERROR_TEXT_SIZE-1-strlen(iso7816_error));

		if (sw2 == 0x82)
			strncat(ptr, "Security status not satisfied.",
				ISO7816_ERROR_TEXT_SIZE-1-strlen(iso7816_error));

		if (sw2 == 0x83)
			strncat(ptr, "Authentication methon blocked.",
				ISO7816_ERROR_TEXT_SIZE-1-strlen(iso7816_error));

		if (sw2 == 0x84)
			strncat(ptr, "Referenced data invalidated.",
				ISO7816_ERROR_TEXT_SIZE-1-strlen(iso7816_error));

		if (sw2 == 0x85)
			strncat(ptr, "Conditions of use not satisfied.",
				ISO7816_ERROR_TEXT_SIZE-1-strlen(iso7816_error));

		if (sw2 == 0x86)
			strncat(ptr, "Command not allowed (no current EF).",
				ISO7816_ERROR_TEXT_SIZE-1-strlen(iso7816_error));

		if (sw2 == 0x87)
			strncat(ptr, "Expected SM data objects missing.",
				ISO7816_ERROR_TEXT_SIZE-1-strlen(iso7816_error));

		if (sw2 == 0x88)
			strncat(ptr, "SM dta objects incorrect.",
				ISO7816_ERROR_TEXT_SIZE-1-strlen(iso7816_error));
	}

	if (sw1 == 0x6A)
	{
		strncat(ptr, "Wrong parameter(s) P1-P2. ",
			ISO7816_ERROR_TEXT_SIZE-strlen(iso7816_error)-1);

		if (sw2 == 0x80)
			strncat(ptr, "Incorrect parameters in the data field.",
				ISO7816_ERROR_TEXT_SIZE-1-strlen(iso7816_error));

		if (sw2 == 0x81)
			strncat(ptr, "Function not supported.",
				ISO7816_ERROR_TEXT_SIZE-1-strlen(iso7816_error));

		if (sw2 == 0x82)
			strncat(ptr, "File not found.",
				ISO7816_ERROR_TEXT_SIZE-1-strlen(iso7816_error));

		if (sw2 == 0x83)
			strncat(ptr, "Record not found.",
				ISO7816_ERROR_TEXT_SIZE-1-strlen(iso7816_error));

		if (sw2 == 0x84)
			strncat(ptr, "Not enough memory space in the file.",
				ISO7816_ERROR_TEXT_SIZE-1-strlen(iso7816_error));

		if (sw2 == 0x85)
			strncat(ptr, "Lc inconsistent with TLV structure.",
				ISO7816_ERROR_TEXT_SIZE-1-strlen(iso7816_error));

		if (sw2 == 0x86)
			strncat(ptr, "Incorrect parameters P1-P2.",
				ISO7816_ERROR_TEXT_SIZE-1-strlen(iso7816_error));

		if (sw2 == 0x87)
			strncat(ptr, "Lc inconsistent with P1-P2.",
				ISO7816_ERROR_TEXT_SIZE-1-strlen(iso7816_error));

		if (sw2 == 0x88)
			strncat(ptr, "Referenced data not found.",
				ISO7816_ERROR_TEXT_SIZE-1-strlen(iso7816_error));
	}
	
	if (sw1 == 0x6B && sw2 == 0x00)
		strncat(ptr, "Wrong parameter(s) P1-P2.",
			ISO7816_ERROR_TEXT_SIZE-strlen(iso7816_error)-1);

	if (sw1 == 0x6C)
#ifdef WIN32
		_snprintf(ptr, ISO7816_ERROR_TEXT_SIZE-strlen(iso7816_error)-1,
			"Wrong length Le: should be %d", sw2);
#else
		snprintf(ptr, ISO7816_ERROR_TEXT_SIZE-strlen(iso7816_error)-1,
			"Wrong length Le: should be %d", sw2);
#endif

	if (sw1 == 0x6D && sw2 == 0x00)
		strncat(ptr,
			"Instruction code not supported or invalid.",
			ISO7816_ERROR_TEXT_SIZE-strlen(iso7816_error)-1);

	if (sw1 == 0x6E && sw2 == 0x00)
		strncat(ptr, "Class not supported.",
			ISO7816_ERROR_TEXT_SIZE-strlen(iso7816_error)-1);

	if (sw1 == 0x6F && sw2 == 0x00)
		strncat(ptr, "No precise diagnosis.",
			ISO7816_ERROR_TEXT_SIZE-strlen(iso7816_error)-1);

	/* default error message */
	if (strlen(iso7816_error) < 9)
		strncat(ptr, "Not defined by 7816.",
			ISO7816_ERROR_TEXT_SIZE-strlen(iso7816_error)-1);

	return iso7816_error;
} /* ISO-7816_ErrorStr */

