/*****************************************************************************
*               Copyright (c) Gemplus Card International 1997
*
* Name        : rftlpreader.c
*
* Description : Card reader interface

  Author      : JL GIRAUD
  Compiler    : Visual C++, linux compiler
  Host        : Windows 9x,NT, Linux
*
* $Id: rftlpreader.c,v 1.4 2000/12/11 17:53:30 rousseau Exp $
* $Log: rftlpreader.c,v $
* Revision 1.4  2000/12/11 17:53:30  rousseau
* Added a 5 seconds timeout if no reader present
*
* Revision 1.3  2000/09/04 15:31:25  giraud
* Change buffer size in tlpreader.* and rftlpreader.*
* Add lib/samples clean-up when make clean is entered in lib/
*
* Revision 1.2  2000/09/04 08:19:06  rousseau
* code cleanup, changed the API and name for [Init|Close]Serial
*
*
******************************************************************************
*
* Warning     :
*
* Remarks     :
*
*****************************************************************************/


#include <fcntl.h>
#include <stdio.h>

#ifndef WIN32
#include <termios.h>
#include <unistd.h>
#endif
#include <errno.h>
#include <string.h>

#include "reader.h"
#include "rftlpreader.h"
#include "bin2c.h"
#include "c2bin.h"
#include "serial.h"

#ifdef WIN32
#define WRITE_COM	WriteCom
#define READ_COM	ReadCom
#else
#define WRITE_COM	write
#define READ_COM	read
#endif

extern int rerrno;


const char RFTLP_RF_OFF[11] = { '6','0','0','2','E','1','1','1',
				'9','2',0x03};
const char RFTLP_RF_ON[11] = { '6','0','0','2','E','1','1','0',
			       '9','3',0x03};
const char RFTLP_GET_CARD[9] = { '6','0','0','1','E','2','8','3',0x03};


const char RFTLP_ACK_BYTE[2] = {'6', '0'};
#define RFTLP_ACK_BYTEh  0x60
const char RFTLP_APDU_EXCH[2] = {'E', '5'};
#define RFTLP_APDU_EXCHh    0xE5

/*****************************************************************************/


int RFTLPResetCardBinary(TYPE_HANDLE fd, unsigned char* atr, int length)
{
	/* Update error code function pointer */
	char buffer[RFTLP_ATR_BUF_SIZE];
	int atr_length, index;

	/* First, PowerDown to make sure the RF is off */
	if ( RFTLPPowerDown(fd) == -1 )
	{
		return -1;
	}

	/* First, try to turn the RF on */

	if (WRITE_COM(fd,RFTLP_RF_ON, sizeof(RFTLP_RF_ON))
		!= sizeof(RFTLP_RF_ON))
	{
		rerrno = RD_WRITE_FAILED;
		return -1;
	}
	/* Get reader response */
	if (RFTLPGetReaderResponse(fd, buffer, sizeof(buffer)) == -1)
	{
		return -1;
	}

	/* Now, let's get a card reset */
	if (WRITE_COM(fd,RFTLP_GET_CARD, sizeof(RFTLP_GET_CARD))
		!= sizeof(RFTLP_GET_CARD))
	{
		rerrno = RD_WRITE_FAILED;
		return -1;
	}

	/* Get reader response */
	atr_length = RFTLPGetReaderResponse(fd, buffer, sizeof(buffer));
	if (atr_length == -1)
	{
		return -1;
	}
	atr_length -= RFTLP_ATR_LENGTH_ADJUST;

	atr_length = (atr_length < length)? atr_length:length;

	for (index = 0; index < atr_length; index++)
		atr[index] = c2bin(buffer+RFTLP_ATR_OFF+2*index,2);

	return atr_length;
}

/*****************************************************************************/


int RFTLPResetCardAscii(TYPE_HANDLE fd, char* atr, int length)
{
	char buffer[RFTLP_ATR_BUF_SIZE];
	int index, atr_length;

	if (length > 0)
		atr[0] = 0;

	/* First, PowerDown to make sure the RF is off */
	if ( RFTLPPowerDown(fd) == -1 )
	{
		return -1;
	}

	/* First, try to turn the RF on */

	if (WRITE_COM(fd,RFTLP_RF_ON, sizeof(RFTLP_RF_ON))
		!= sizeof(RFTLP_RF_ON))
	{
		rerrno = RD_WRITE_FAILED;
		return -1;
	}
	/* Get reader response */
	if (RFTLPGetReaderResponse(fd, buffer, sizeof(buffer)) == -1)
	{
		return -1;
	}

	/* Now, let's get a card reset */
	if (WRITE_COM(fd,RFTLP_GET_CARD, sizeof(RFTLP_GET_CARD))
		!= sizeof(RFTLP_GET_CARD))
	{
		rerrno = RD_WRITE_FAILED;
		return -1;
	}

	/* Get reader response (minus ATR length)*/
	atr_length = RFTLPGetReaderResponse(fd, buffer, sizeof(buffer));
	if (atr_length == -1)
	{
		return -1;
	}

	/* Compute ASCII length */
	length -= 1; /* Count NULL byte */
	atr_length -= RFTLP_ATR_LENGTH_ADJUST;
	atr_length <<=  1;
	atr_length = (atr_length < length)? atr_length:length;

	for (index = 0; index < atr_length; index++)
	atr[index] = buffer[RFTLP_ATR_OFF+index];

	/* Add NULL byte */
	atr[index] = '\0';

	return atr_length+1;
}

/*****************************************************************************/


int RFTLPGetReaderResponse(TYPE_HANDLE fd, char *buffer, int length)
{
	int index;
	int ack_byte;
	t_rd_error error;

	index = 0;
	do
	{
		if (READ_COM(fd,&buffer[index++],1) !=1)
		{
			rerrno = RD_READ_ERROR;
			return -1;
		}
		if (index >= length)
		{
			rerrno = RD_RESP_TOO_LONG;
			return -1;
		}
	}
	while (buffer[index-1] != RFTLP_EOT) ;

	ack_byte = c2bin(buffer+RFTLP_ACK_OFF,2);
	if (ack_byte != RFTLP_TRANS_OK)
	{
		rerrno = RD_TRANS_ERROR;
		return -1;
	}

	/* Check return code (OK or errors) */
	error = RFTLPTrError(c2bin(buffer+RFTLP_CMD_OFF,2));
	if ( error != RD_NO_ERROR )
	{
		/* The reader returned an error */
		rerrno = error;
		return -1;
	}

	/* return length - 1 for CMD byte */
	return c2bin(buffer+RFTLP_LENGTH_OFF,2)-1;
}





/*****************************************************************************/




t_rd_error RFTLPTrError(int reader_error)
{
	t_rd_error error;

	switch(reader_error)
	{
		case RFTLP_TRANSFERT_TO_CARD_INTERRUPTED:
		case RFTLP_CARD_RETURNED_ERROR:
		case RFTLP_NO_ERROR:
			error = RD_NO_ERROR;
			break;

		case RFTLP_UNKNOWN_COMMAND:
		case RFTLP_COMMAND_UNKNOWN:
			error = RD_UNKNOWN_COMMAND;
			break;

		case RFTLP_DRIVER_LIMITATION:
			error = RD_DRIVER_LIMITATION;
			break;

		case RFTLP_WRONG_ARG_LENGTH:
			error = RD_WRONG_ARG_LENGTH;
			break;

		case RFTLP_PROTOCOL_ERROR:
			error = RD_PROTOCOL_ERROR;
			break;

		case RFTLP_WRONG_ATR:
			error =  RD_WRONG_ATR;
			break;

		case RFTLP_WRONG_RD_INS:
			error = RD_WRONG_RD_INS;
			break;

		case RFTLP_MSG_TOO_LONG:
			error = RD_MSG_TOO_LONG;
			break;

		case RFTLP_READ_ERROR:
			error = RD_READ_ERROR;
			break;

		case RFTLP_CARD_OFF:
			error = RD_CARD_OFF;
			break;

		case RFTLP_PROG_VOLT_ERROR:
			error = RD_PROG_VOLT_ERROR;
			break;

		case RFTLP_UNKNOWN_PROTOCOL2:
		case RFTLP_UNKNOWN_PROTOCOL1:
			error = RD_UNKNOWN_PROTOCOL;
			break;

		case RFTLP_LENGTH_ERROR:
			error = RD_LENGTH_ERROR;
			break;

		case RFTLP_WRONG_DATA:
		case RFTLP_PROTECTION_ERROR:
		case RFTLP_TCK_ERROR:
		case RFTLP_ILLEGAL_ACCESS:
			error = RD_USELESS_ERROR;
			break;

		case RFTLP_WRONG_NB_OF_PARAM:
			error = RD_WRONG_NB_OF_PARAM;
			break;

		case RFTLP_UNKNOWN_CARD_PROTOCOL:
			error = RD_UNKNOWN_CARD_PROTOCOL;
			break;

		case RFTLP_CARD_MUTE:
			error = RD_CARD_MUTE;
			break;

		case RFTLP_CARD_PARITY_ERROR:
			error = RD_CARD_PARITY_ERROR;
			break;

		case RFTLP_CARD_WRONG_PROCEDURE_BYTE:
			error = RD_CARD_WRONG_PROCEDURE_BYTE;
			break;

		case RFTLP_CARD_REMOVED:
			error = RD_CARD_REMOVED;
			break;

		case RFTLP_SHORT_CIRCUIT:
			error = RD_SHORT_CIRCUIT;
			break;

		case RFTLP_CARD_ABSENT:
			error = RD_CARD_ABSENT;
			break;
		
		default:
			error = RD_UNKNOWN_ERROR;
	}

	return error;
}




/*****************************************************************************/




int RFTLPPowerDown(TYPE_HANDLE fd)
{
	char buffer[RFTLP_PW_DOWN_BUF_SIZE];
	int index;

	if (WRITE_COM(fd,RFTLP_RF_OFF, sizeof(RFTLP_RF_OFF))
	!= sizeof(RFTLP_RF_OFF))
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}
	/* Get reader response */
	index = RFTLPGetReaderResponse(fd, buffer, sizeof(buffer));
	return (index==-1)?index:0;;
}





/*****************************************************************************/





int RFTLPSendIncomingAPDUBinary(TYPE_HANDLE fd,
				unsigned char CLA,
				unsigned char INS,
				unsigned char P1,
				unsigned char P2,
				unsigned char L,
				const unsigned char *data_in,
				unsigned char *ad_SW1,
				unsigned char *ad_SW2)
{
	int cks_val;
	int data_length;
	char numb_buf[2];
	int i;
	int sw_length;
	char buffer[RFTLP_SW_BUF_SIZE];
	unsigned char rd_EOT;


	rd_EOT = RFTLP_EOT;
	data_length = L+5;  /* APDU size + data */
	data_length++; /* CMD byte */

	/* Init checksum value */
	cks_val = RFTLP_ACK_BYTEh ^ data_length ^ RFTLP_APDU_EXCHh;

	if (WRITE_COM(fd,RFTLP_ACK_BYTE, sizeof(RFTLP_ACK_BYTE))
	!= sizeof(RFTLP_ACK_BYTE))
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}

	bin2c(data_length, numb_buf, 2);
	if (WRITE_COM(fd,numb_buf, 2) != 2)
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}

	if (WRITE_COM(fd,RFTLP_APDU_EXCH, sizeof(RFTLP_APDU_EXCH))
	!= sizeof(RFTLP_APDU_EXCH))
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}

	bin2c(CLA, numb_buf, 2);
	if (WRITE_COM(fd,numb_buf, 2) != 2)
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}
	cks_val ^= CLA;

	bin2c(INS, numb_buf, 2);
	if (WRITE_COM(fd,numb_buf, 2) != 2)
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}
	cks_val ^= INS;

	bin2c(P1, numb_buf, 2);
	if (WRITE_COM(fd,numb_buf, 2) != 2)
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}
	cks_val ^= P1;

	bin2c(P2, numb_buf, 2);
	if (WRITE_COM(fd,numb_buf, 2) != 2)
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}
	cks_val ^= P2;

	bin2c(L, numb_buf, 2);
	if (WRITE_COM(fd,numb_buf, 2) != 2)
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}
	cks_val ^= L;

	for (i=0; i < L; i++)
	{
	bin2c(data_in[i], numb_buf, 2);
	if (WRITE_COM(fd,numb_buf, 2) != 2)
	{
	    rerrno = RD_WRITE_FAILED;
	    return -1;
	}
	cks_val ^= data_in[i];
	}

	bin2c(cks_val, numb_buf, 2);
	if (WRITE_COM(fd,numb_buf, 2) != 2)
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}

	/* Send EOT to mark the end of the message */
	if (WRITE_COM(fd,&rd_EOT, 1) != 1)
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}


	/* Get card response */
	sw_length = RFTLPGetReaderResponse(fd, buffer, sizeof(buffer));
	if (sw_length == -1)
	{
	return -1;
	}
	if (sw_length != 2)
	{
	rerrno =  RD_STATUS_TOO_LONG;
	return -1;
	}

	*ad_SW1 = c2bin(buffer+RFTLP_DATA_OFF,2);
	*ad_SW2 = c2bin(buffer+RFTLP_DATA_OFF+2,2);
	return 2;
}




/*****************************************************************************/




int RFTLPSendOutgoingAPDUBinary(TYPE_HANDLE fd,
				unsigned char CLA,
				unsigned char INS,
				unsigned char P1,
				unsigned char P2,
				unsigned char L,
				unsigned char *data_out,
				unsigned char *ad_SW1,
				unsigned char *ad_SW2)
{
	int cks_val;
	int data_length;
	char numb_buf[2];
	int i;
	int resp_length;
	char buffer[RFTLP_DATA_BUF_SIZE];
	unsigned char rd_EOT;


	rd_EOT = RFTLP_EOT;
	data_length = 5;  /* APDU size + data */
	data_length++;    /* CMD byte */

	/* Init checksum value */
	cks_val = RFTLP_ACK_BYTEh ^ data_length ^ RFTLP_APDU_EXCHh;

	if (WRITE_COM(fd,RFTLP_ACK_BYTE, sizeof(RFTLP_ACK_BYTE))
	!= sizeof(RFTLP_ACK_BYTE))
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}

	bin2c(data_length, numb_buf, 2);
	if (WRITE_COM(fd,numb_buf, 2) != 2)
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}

	if (WRITE_COM(fd,RFTLP_APDU_EXCH, sizeof(RFTLP_APDU_EXCH))
	!= sizeof(RFTLP_APDU_EXCH))
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}

	bin2c(CLA, numb_buf, 2);
	if (WRITE_COM(fd,numb_buf, 2) != 2)
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}
	cks_val ^= CLA;

	bin2c(INS, numb_buf, 2);
	if (WRITE_COM(fd,numb_buf, 2) != 2)
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}
	cks_val ^= INS;

	bin2c(P1, numb_buf, 2);
	if (WRITE_COM(fd,numb_buf, 2) != 2)
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}
	cks_val ^= P1;

	bin2c(P2, numb_buf, 2);
	if (WRITE_COM(fd,numb_buf, 2) != 2)
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}
	cks_val ^= P2;

	bin2c(L, numb_buf, 2);
	if (WRITE_COM(fd,numb_buf, 2) != 2)
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}
	cks_val ^= L;

	bin2c(cks_val, numb_buf, 2);
	if (WRITE_COM(fd,numb_buf, 2) != 2)
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}

	/* Send EOT to mark the end of the message */
	if (WRITE_COM(fd,&rd_EOT, 1) != 1)
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}


	/* Get card response */
	resp_length = RFTLPGetReaderResponse(fd, buffer, sizeof(buffer));
	if (resp_length == -1)
	{
	return -1;
	}

	resp_length -= 2; /* SW is not part of data */
	resp_length = (resp_length < L)? resp_length:L;

	for (i = 0; i < resp_length; i++)
	{
	data_out[i] = c2bin(buffer+RFTLP_DATA_OFF+2*i,2);
	}
	
	*ad_SW1 = c2bin(buffer+RFTLP_DATA_OFF+2*resp_length,2);
	*ad_SW2 = c2bin(buffer+RFTLP_DATA_OFF+2*resp_length+2,2);
	return resp_length;
}







/*****************************************************************************/





int RFTLPSendIncomingOutgoingAPDUBinary(TYPE_HANDLE fd,
					unsigned char CLA,
					unsigned char INS,
					unsigned char P1,
					unsigned char P2,
					unsigned char L_in,
					const unsigned char *data_in,
					unsigned char L_out,
					unsigned char *data_out,
					unsigned char *ad_SW1,
					unsigned char *ad_SW2)
{
	int cks_val;
	int data_length;
	char numb_buf[2];
	int i;
	int length;
	int resp_length;
	char buffer[RFTLP_DATA_BUF_SIZE];
	unsigned char rd_EOT;


	rd_EOT = RFTLP_EOT;
	data_length = L_in+6;  /* APDU size + data + Le */
	data_length++; /* CMD byte */

	/* Init checksum value */
	cks_val = RFTLP_ACK_BYTEh ^ data_length ^ RFTLP_APDU_EXCHh;

	if (WRITE_COM(fd,RFTLP_ACK_BYTE, sizeof(RFTLP_ACK_BYTE))
	!= sizeof(RFTLP_ACK_BYTE))
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}

	bin2c(data_length, numb_buf, 2);
	if (WRITE_COM(fd,numb_buf, 2) != 2)
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}

	if (WRITE_COM(fd,RFTLP_APDU_EXCH, sizeof(RFTLP_APDU_EXCH))
	!= sizeof(RFTLP_APDU_EXCH))
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}

	bin2c(CLA, numb_buf, 2);
	if (WRITE_COM(fd,numb_buf, 2) != 2)
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}
	cks_val ^= CLA;

	bin2c(INS, numb_buf, 2);
	if (WRITE_COM(fd,numb_buf, 2) != 2)
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}
	cks_val ^= INS;

	bin2c(P1, numb_buf, 2);
	if (WRITE_COM(fd,numb_buf, 2) != 2)
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}
	cks_val ^= P1;

	bin2c(P2, numb_buf, 2);
	if (WRITE_COM(fd,numb_buf, 2) != 2)
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}
	cks_val ^= P2;

	bin2c(L_in, numb_buf, 2);
	if (WRITE_COM(fd,numb_buf, 2) != 2)
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}
	cks_val ^= L_in;

	for (i=0; i < L_in; i++)
	{
	bin2c(data_in[i], numb_buf, 2);
	if (WRITE_COM(fd,numb_buf, 2) != 2)
	{
	    rerrno = RD_WRITE_FAILED;
	    return -1;
	}
	cks_val ^= data_in[i];
	}

	bin2c(L_out, numb_buf, 2);
	if (WRITE_COM(fd,numb_buf, 2) != 2)
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}
	cks_val ^= L_out;

	bin2c(cks_val, numb_buf, 2);
	if (WRITE_COM(fd,numb_buf, 2) != 2)
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}

	/* Send EOT to mark the end of the message */
	if (WRITE_COM(fd,&rd_EOT, 1) != 1)
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}


	/* Get card response */
	resp_length = RFTLPGetReaderResponse(fd, buffer, sizeof(buffer));
	if (resp_length == -1)
	{
	return -1;
	}
	resp_length -= 2; /* SW is not part of data */
	length = (resp_length < L_out)? resp_length:L_out;

	for (i = 0; i < length; i++)
	{
	data_out[i] = c2bin(buffer+RFTLP_DATA_OFF+2*i,2);
	}
	
	*ad_SW1 = c2bin(buffer+RFTLP_DATA_OFF+2*resp_length,2);
	*ad_SW2 = c2bin(buffer+RFTLP_DATA_OFF+2*resp_length+2,2);
	return length;

}




/*****************************************************************************/



int RFTLPSendIncomingAPDUAscii(TYPE_HANDLE fd, const char* APDU_Data,
			       char* ad_response, int resp_length)
{
	int cks_val;
	int data_length;
	char numb_buf[2];
	int i;
	int sw_length;
	char buffer[RFTLP_SW_BUF_SIZE];
	unsigned char rd_EOT;


	rd_EOT = RFTLP_EOT;
	/* 2 chars per Byte */
	data_length = strlen(APDU_Data) >> 1;
	data_length++; /* CMD byte */

	/* Checksum value */
	cks_val = RFTLP_ACK_BYTEh ^ data_length ^ RFTLP_APDU_EXCHh;
	for (i=0; i<(data_length-1); i++)
	{
	cks_val ^= c2bin(APDU_Data+2*i,2);
	}

	if (WRITE_COM(fd,RFTLP_ACK_BYTE, sizeof(RFTLP_ACK_BYTE))
	!= sizeof(RFTLP_ACK_BYTE))
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}

	bin2c(data_length, numb_buf, 2);
	if (WRITE_COM(fd,numb_buf, 2) != 2)
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}

	if (WRITE_COM(fd,RFTLP_APDU_EXCH, sizeof(RFTLP_APDU_EXCH))
	!= sizeof(RFTLP_APDU_EXCH))
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}

	data_length--;
	data_length <<= 1;
	if (WRITE_COM(fd,APDU_Data, data_length)
	!= data_length )
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}

	bin2c(cks_val, numb_buf, 2);
	if (WRITE_COM(fd,numb_buf, 2) != 2)
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}

	/* Send EOT to mark the end of the message */
	if (WRITE_COM(fd,&rd_EOT, 1) != 1)
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}


	/* Get card response */
	sw_length = RFTLPGetReaderResponse(fd, buffer, sizeof(buffer));
	if (sw_length == -1)
	{
	return -1;
	}

	/* Count in chars  */
	sw_length <<= 1;
	/* Add count for '\0' */
	sw_length++;

	resp_length = (sw_length<resp_length)?sw_length:resp_length;
	if (resp_length > 0)
	{
	memcpy(ad_response,buffer+RFTLP_DATA_OFF,resp_length-1);
	ad_response[resp_length-1] = '\0';
	}
	return resp_length;
}





/*****************************************************************************/



int RFTLPSendOutgoingAPDUAscii(TYPE_HANDLE fd, const char* APDU_Data,
			       char* ad_response, int resp_length)
{
	int cks_val;
	int data_length;
	char numb_buf[2];
	int i;
	int sw_length;
	char buffer[RFTLP_DATA_BUF_SIZE];
	unsigned char rd_EOT;


	rd_EOT = RFTLP_EOT;
	/* 2 chars per Byte */
	data_length = 5;  /* APDU bytes */
	data_length++;    /* CMD byte   */

	/* Checksum value */
	cks_val = RFTLP_ACK_BYTEh ^ data_length ^ RFTLP_APDU_EXCHh;
	for (i=0; i<(data_length-1); i++)
	{
	cks_val ^= c2bin(APDU_Data+2*i,2);
	}

	if (WRITE_COM(fd,RFTLP_ACK_BYTE, sizeof(RFTLP_ACK_BYTE))
	!= sizeof(RFTLP_ACK_BYTE))
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}

	bin2c(data_length, numb_buf, 2);
	if (WRITE_COM(fd,numb_buf, 2) != 2)
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}

	if (WRITE_COM(fd,RFTLP_APDU_EXCH, sizeof(RFTLP_APDU_EXCH))
	!= sizeof(RFTLP_APDU_EXCH))
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}

	if (WRITE_COM(fd,APDU_Data, 10)
	!= 10 )
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}


	bin2c(cks_val, numb_buf, 2);
	if (WRITE_COM(fd,numb_buf, 2) != 2)
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}

	/* Send EOT to mark the end of the message */
	if (WRITE_COM(fd,&rd_EOT, 1) != 1)
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}


	/* Get card response */
	sw_length = RFTLPGetReaderResponse(fd, buffer, sizeof(buffer));
	if (sw_length == -1)
	{
	return -1;
	}

	/* Count in chars  */
	sw_length <<= 1;
	/* Add count for '\0' */
	sw_length++;

	resp_length = (sw_length<resp_length)?sw_length:resp_length;
	if (resp_length > 0)
	{
	memcpy(ad_response,buffer+RFTLP_DATA_OFF, resp_length-1);
	ad_response[resp_length-1] = '\0';
	}
	return resp_length;

}




/*****************************************************************************/



int RFTLPSendIncomingOutgoingAPDUAscii(TYPE_HANDLE fd,
				       const char* APDU_Data,
				       char* ad_response, int resp_length)
{
	int cks_val;
	int data_length;
	char numb_buf[2];
	int i;
	int length;
	char buffer[RFTLP_DATA_BUF_SIZE];
	unsigned char rd_EOT;


	rd_EOT = RFTLP_EOT;
	/* 2 chars per Byte */
	data_length = strlen(APDU_Data) >> 1;
	data_length++; /* CMD byte */
	data_length++; /* Le byte  */

	/* Expected length in bytes (>> gets READ_COM of the '\0' count) */
	length = resp_length >> 1 ;
	/* actual length is length - 2 bytes for the SW */
	length -= 2;
	
	
	/* Checksum value */
	cks_val = RFTLP_ACK_BYTEh ^ data_length ^ RFTLP_APDU_EXCHh;

	/* substract 2 (instead of 1) because data_length includes Le */
	for (i=0; i<(data_length-2); i++)
	{
	cks_val ^= c2bin(APDU_Data+2*i,2);
	}
	
	/* For this reader, expected length is sent after the data */
	cks_val ^= length;

	if (WRITE_COM(fd,RFTLP_ACK_BYTE, sizeof(RFTLP_ACK_BYTE))
	!= sizeof(RFTLP_ACK_BYTE))
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}

	bin2c(data_length, numb_buf, 2);
	if (WRITE_COM(fd,numb_buf, 2) != 2)
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}

	if (WRITE_COM(fd,RFTLP_APDU_EXCH, sizeof(RFTLP_APDU_EXCH))
	!= sizeof(RFTLP_APDU_EXCH))
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}

	data_length--; /* Remove count for CMD byte */
	data_length--; /* Remove count for Le       */
	data_length <<= 1;
	if (WRITE_COM(fd,APDU_Data, data_length)
	!= data_length )
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}

	bin2c(length, numb_buf, 2);
	if (WRITE_COM(fd,numb_buf, 2) != 2)
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}


	bin2c(cks_val, numb_buf, 2);
	if (WRITE_COM(fd,numb_buf, 2) != 2)
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}

	/* Send EOT to mark the end of the message */
	if (WRITE_COM(fd,&rd_EOT, 1) != 1)
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}


	/* Get card response */
	length = RFTLPGetReaderResponse(fd, buffer, sizeof(buffer));
	if (length == -1)
	{
	return -1;
	}

	/* Count in chars  */
	length <<= 1;
	/* Add count for '\0' */
	length++;


	resp_length = (length<resp_length)?length:resp_length;
	if (resp_length > 0)
	{
	memcpy(ad_response,buffer+RFTLP_DATA_OFF,resp_length-1);
	ad_response[resp_length-1] = '\0';
	}
	return resp_length;
}
