/*****************************************************************************
*               Copyright (c) Gemplus Card International 1997
*
* Name        : tlpreader.c
*
* Description : Card reader interface

  Author      : JL GIRAUD
  Compiler    : Visual C++, linux compiler
  Host        : Windows 9x,NT, Linux
*
* $Id: tlpreader.c,v 1.4 2000/12/11 17:53:30 rousseau Exp $
* $Log: tlpreader.c,v $
* Revision 1.4  2000/12/11 17:53:30  rousseau
* Added a 5 seconds timeout if no reader present
*
* Revision 1.3  2000/09/04 15:31:25  giraud
* Change buffer size in tlpreader.* and rftlpreader.*
* Add lib/samples clean-up when make clean is entered in lib/
*
* Revision 1.2  2000/09/04 08:19:06  rousseau
* code cleanup, changed the API and name for [Init|Close]Serial
*
*
******************************************************************************
*
* Warning     :
*
* Remarks     :
*
*****************************************************************************/


#include <fcntl.h>

#ifndef WIN32
#include <termios.h>
#include <unistd.h>
#endif

#include <stdio.h>
#include <errno.h>
#include <string.h>

#include "reader.h"
#include "tlpreader.h"
#include "bin2c.h"
#include "c2bin.h"
#include "serial.h"

#ifdef WIN32
#define WRITE_COM	WriteCom
#define READ_COM	ReadCom
#else
#define WRITE_COM	write
#define READ_COM	read_with_timeout
#endif

extern int rerrno;


const char TLP_SET_MODE[] =    { 0x01,0x00};
const char TLP_POWER_DOWN[9] = { '6','0','0','1','4','D','2','C',0x03};
const char TLP_POWER_UP[15] =  { '6','0','0','4','6','E','0','0',
				'0','0', '0','0','0','A',0x03};

const char TLP_GTR_CMD[] = { '6', '0', '0', '6', 'D', 'B', 'O', '0', 'C',
			     '0', '0', '0', '0', '0'};
#define TLP_GTR_CMD_CKS 0x7D

const char TLP_ACK_BYTE[2] = {'6', '0'};
#define TLP_ACK_BYTEh  0x60
const char TLP_IN_CMD[2] = {'D', 'A'};
#define TLP_IN_CMDh    0xDA
const char TLP_OUT_CMD[2] = {'D', 'B'};
#define TLP_OUT_CMDh    0xDB

#define TLP_GTR_REQ_SW1 0x61
#define TLP_GTR_REQ_SW1_C1 '6'
#define TLP_GTR_REQ_SW1_C2 '1'

/****************************************************************************/


int TLPResetCardBinary(TYPE_HANDLE fd, unsigned char* atr, int length)
{
	/* Update error code function pointer */
	char buffer[TLP_ATR_BUF_SIZE];
	int atr_length, index;

	/* First, try TLP reset */

	if (WRITE_COM(fd,TLP_POWER_UP, sizeof(TLP_POWER_UP))
		!= sizeof(TLP_POWER_UP))
	{
		rerrno = RD_WRITE_FAILED;
		return -1;
	}
	/* Get reader response */
	atr_length = TLPGetReaderResponse(fd, buffer, sizeof(buffer));
	if (atr_length == -1)
		return -1;

	atr_length -= TLP_ATR_LENGTH_ADJUST;

	atr_length = (atr_length < length)? atr_length:length;

	for (index = 0; index < atr_length; index++)
		atr[index] = c2bin(buffer+TLP_ATR_OFF+2*index, 2);

	return atr_length;
}

/*****************************************************************************/


int TLPResetCardAscii(TYPE_HANDLE fd, char* atr, int length)
{
	char buffer[TLP_ATR_BUF_SIZE];
	int index, atr_length;

	/* first empty the buffer */
	if (length > 0)
	atr[0] = '\0';

	/* First, try TLP reset */

	if (WRITE_COM(fd,TLP_POWER_UP, sizeof(TLP_POWER_UP))
		!= sizeof(TLP_POWER_UP))
	{
		rerrno = RD_READ_ERROR;
		return -1;
	}
	
	/* Get reader response (minus ATR length)*/
	atr_length = TLPGetReaderResponse(fd, buffer, sizeof(buffer));
	if (atr_length == -1)
		return -1;

	/* Compute ASCII length */
	length -= 1; /* Count NULL byte */
	atr_length -= TLP_ATR_LENGTH_ADJUST;
	atr_length <<=  1;
	atr_length = (atr_length < length) ? atr_length : length;

	for (index = 0; index < atr_length; index++)
		atr[index] = buffer[TLP_ATR_OFF+index];

	/* Add NULL byte */
	atr[index] = '\0';

	return atr_length;
}

/*****************************************************************************/


int TLPGetReaderResponse(TYPE_HANDLE fd, char *buffer, int length)
{
	int index;
	int ack_byte;
	t_rd_error error;

	index = 0;
	do
	{
		if (READ_COM(fd,&buffer[index++],1) !=1)
		{
			rerrno = RD_READ_ERROR;
			return -1;
		}
		if (index >= length)
		{
			rerrno = RD_RESP_TOO_LONG;
			return -1;
		}
	}
	while (buffer[index-1] != TLP_EOT) ;

	ack_byte = c2bin(buffer+TLP_ACK_OFF,2);
	if (ack_byte != TLP_TRANS_OK)
	{
		rerrno = RD_TRANS_ERROR;
		return -1;
	}

	/* Check return code (OK or errors) */
	error = TLPTrError(c2bin(buffer+TLP_CMD_OFF,2));
	if ( error != RD_NO_ERROR )
	{
		/* The reader returned an error */
		rerrno = error;
		return -1;
	}

	/* return length - 1 for CMD byte */
	return c2bin(buffer+TLP_LENGTH_OFF,2)-1;
}

/*****************************************************************************/


t_rd_error TLPTrError(int reader_error)
{
	t_rd_error error;

	switch(reader_error)
	{
		case TLP_TRANSFERT_TO_CARD_INTERRUPTED:
		case TLP_CARD_RETURNED_ERROR:
		case TLP_NO_ERROR:
			error = RD_NO_ERROR;
			break;

		case TLP_UNKNOWN_COMMAND:
		case TLP_COMMAND_UNKNOWN:
			error = RD_UNKNOWN_COMMAND;
			break;

		case TLP_DRIVER_LIMITATION:
			error = RD_DRIVER_LIMITATION;
			break;

		case TLP_WRONG_ARG_LENGTH:
			error = RD_WRONG_ARG_LENGTH;
			break;

		case TLP_PROTOCOL_ERROR:
			error = RD_PROTOCOL_ERROR;
			break;

		case TLP_WRONG_ATR:
			error =  RD_WRONG_ATR;
			break;

		case TLP_WRONG_RD_INS:
			error = RD_WRONG_RD_INS;
			break;

		case TLP_MSG_TOO_LONG:
			error = RD_MSG_TOO_LONG;
			break;

		case TLP_READ_ERROR:
			error = RD_READ_ERROR;
			break;

		case TLP_CARD_OFF:
			error = RD_CARD_OFF;
			break;

		case TLP_PROG_VOLT_ERROR:
			error = RD_PROG_VOLT_ERROR;
			break;

		case TLP_UNKNOWN_PROTOCOL2:
		case TLP_UNKNOWN_PROTOCOL1:
			error = RD_UNKNOWN_PROTOCOL;
			break;

		case TLP_LENGTH_ERROR:
			error = RD_LENGTH_ERROR;
			break;

		case TLP_WRONG_DATA:
		case TLP_PROTECTION_ERROR:
		case TLP_TCK_ERROR:
		case TLP_ILLEGAL_ACCESS:
			error = RD_USELESS_ERROR;
			break;

		case TLP_WRONG_NB_OF_PARAM:
			error = RD_WRONG_NB_OF_PARAM;
			break;

		case TLP_UNKNOWN_CARD_PROTOCOL:
			error = RD_UNKNOWN_CARD_PROTOCOL;
			break;

		case TLP_CARD_MUTE:
			error = RD_CARD_MUTE;
			break;

		case TLP_CARD_PARITY_ERROR:
			error = RD_CARD_PARITY_ERROR;
			break;

		case TLP_CARD_WRONG_PROCEDURE_BYTE:
			error = RD_CARD_WRONG_PROCEDURE_BYTE;
			break;

		case TLP_CARD_REMOVED:
			error = RD_CARD_REMOVED;
			break;

		case TLP_SHORT_CIRCUIT:
			error = RD_SHORT_CIRCUIT;
			break;

		case TLP_CARD_ABSENT:
			error = RD_CARD_ABSENT;
			break;

		default:
			error = RD_UNKNOWN_ERROR;
	}

	return error;
}

/*****************************************************************************/


int TLPPowerDown(TYPE_HANDLE fd)
{
	char buffer[TLP_PW_DOWN_BUF_SIZE];
	int index;

	if (WRITE_COM(fd,TLP_POWER_DOWN, sizeof(TLP_POWER_DOWN))
	!= sizeof(TLP_POWER_DOWN))
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}
	/* Get reader response */
	index = TLPGetReaderResponse(fd, buffer, sizeof(buffer));
	return (index==-1)?index:0;;
}

/*****************************************************************************/


int TLPSendIncomingAPDUBinary(TYPE_HANDLE fd,
	unsigned char CLA,
	unsigned char INS,
	unsigned char P1,
	unsigned char P2,
	unsigned char L,
	const unsigned char *data_in,
	unsigned char *ad_SW1,
	unsigned char *ad_SW2)
{
	int cks_val;
	int data_length;
	char numb_buf[2];
	int i;
	int sw_length;
	char buffer[TLP_SW_BUF_SIZE];
	unsigned char rd_EOT;

	rd_EOT = TLP_EOT;
	data_length = L+5;  /* APDU size + data */
	data_length++; /* CMD byte */

	/* Init checksum value */
	cks_val = TLP_ACK_BYTEh ^ data_length ^ TLP_IN_CMDh;

	if (WRITE_COM(fd,TLP_ACK_BYTE, sizeof(TLP_ACK_BYTE))
	!= sizeof(TLP_ACK_BYTE))
	{
		rerrno = RD_WRITE_FAILED;
		return -1;
	}

	bin2c(data_length, numb_buf, 2);
	if (WRITE_COM(fd,numb_buf, 2) != 2)
	{
		rerrno = RD_WRITE_FAILED;
		return -1;
	}

	if (WRITE_COM(fd,TLP_IN_CMD, sizeof(TLP_IN_CMD))
	!= sizeof(TLP_IN_CMD))
	{
		rerrno = RD_WRITE_FAILED;
		return -1;
	}

	bin2c(CLA, numb_buf, 2);
	if (WRITE_COM(fd,numb_buf, 2) != 2)
	{
		rerrno = RD_WRITE_FAILED;
		return -1;
	}
	cks_val ^= CLA;

	bin2c(INS, numb_buf, 2);
	if (WRITE_COM(fd,numb_buf, 2) != 2)
	{
		rerrno = RD_WRITE_FAILED;
		return -1;
	}
	cks_val ^= INS;

	bin2c(P1, numb_buf, 2);
	if (WRITE_COM(fd,numb_buf, 2) != 2)
	{
		rerrno = RD_WRITE_FAILED;
		return -1;
	}
	cks_val ^= P1;

	bin2c(P2, numb_buf, 2);
	if (WRITE_COM(fd,numb_buf, 2) != 2)
	{
		rerrno = RD_WRITE_FAILED;
		return -1;
	}
	cks_val ^= P2;

	bin2c(L, numb_buf, 2);
	if (WRITE_COM(fd,numb_buf, 2) != 2)
	{
		rerrno = RD_WRITE_FAILED;
		return -1;
	}
	cks_val ^= L;

	for (i=0; i < L; i++)
	{
	bin2c(data_in[i], numb_buf, 2);
	if (WRITE_COM(fd,numb_buf, 2) != 2)
	{
	    rerrno = RD_WRITE_FAILED;
	    return -1;
	}
	cks_val ^= data_in[i];
	}

	bin2c(cks_val, numb_buf, 2);
	if (WRITE_COM(fd,numb_buf, 2) != 2)
	{
		rerrno = RD_WRITE_FAILED;
		return -1;
	}

	/* Send EOT to mark the end of the message */
	if (WRITE_COM(fd,&rd_EOT, 1) != 1)
	{
		rerrno = RD_WRITE_FAILED;
		return -1;
	}

	/* Get card response */
	sw_length = TLPGetReaderResponse(fd, buffer, sizeof(buffer));
	if (sw_length == -1)
	{
		rerrno = RD_UNKNOWN_ERROR;	/* todo: find the good error message */
		return -1;
	}

	if (sw_length != 2)
	{
		rerrno =  RD_STATUS_TOO_LONG;
		return -1;
	}

	*ad_SW1 = c2bin(buffer+TLP_DATA_OFF,2);
	*ad_SW2 = c2bin(buffer+TLP_DATA_OFF+2,2);

	return 2;
}




/*****************************************************************************/




int TLPSendOutgoingAPDUBinary(TYPE_HANDLE fd,
			      unsigned char CLA,
			      unsigned char INS,
			      unsigned char P1,
			      unsigned char P2,
			      unsigned char L,
			      unsigned char *data_out,
			      unsigned char *ad_SW1,
			      unsigned char *ad_SW2)
{
	int cks_val;
	int data_length;
	char numb_buf[2];
	int i;
	int resp_length;
	char buffer[TLP_DATA_BUF_SIZE];
	unsigned char rd_EOT;


	rd_EOT = TLP_EOT;
	data_length = 5;  /* APDU size + data */
	data_length++;    /* CMD byte */

	/* Init checksum value */
	cks_val = TLP_ACK_BYTEh ^ data_length ^ TLP_OUT_CMDh;

	if (WRITE_COM(fd,TLP_ACK_BYTE, sizeof(TLP_ACK_BYTE))
	!= sizeof(TLP_ACK_BYTE))
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}

	bin2c(data_length, numb_buf, 2);
	if (WRITE_COM(fd,numb_buf, 2) != 2)
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}

	if (WRITE_COM(fd,TLP_OUT_CMD, sizeof(TLP_OUT_CMD))
	!= sizeof(TLP_OUT_CMD))
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}

	bin2c(CLA, numb_buf, 2);
	if (WRITE_COM(fd,numb_buf, 2) != 2)
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}
	cks_val ^= CLA;

	bin2c(INS, numb_buf, 2);
	if (WRITE_COM(fd,numb_buf, 2) != 2)
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}
	cks_val ^= INS;

	bin2c(P1, numb_buf, 2);
	if (WRITE_COM(fd,numb_buf, 2) != 2)
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}
	cks_val ^= P1;

	bin2c(P2, numb_buf, 2);
	if (WRITE_COM(fd,numb_buf, 2) != 2)
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}
	cks_val ^= P2;

	bin2c(L, numb_buf, 2);
	if (WRITE_COM(fd,numb_buf, 2) != 2)
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}
	cks_val ^= L;

	bin2c(cks_val, numb_buf, 2);
	if (WRITE_COM(fd,numb_buf, 2) != 2)
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}

	/* Send EOT to mark the end of the message */
	if (WRITE_COM(fd,&rd_EOT, 1) != 1)
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}


	/* Get card response */
	resp_length = TLPGetReaderResponse(fd, buffer, sizeof(buffer));
	if (resp_length == -1)
	{
	return -1;
	}

	resp_length -= 2; /* SW is not part of data */
	resp_length = (resp_length < L)? resp_length:L;

	for (i = 0; i < resp_length; i++)
	{
	data_out[i] = c2bin(buffer+TLP_DATA_OFF+2*i,2);
	}
	
	*ad_SW1 = c2bin(buffer+TLP_DATA_OFF+2*resp_length,2);
	*ad_SW2 = c2bin(buffer+TLP_DATA_OFF+2*resp_length+2,2);
	return resp_length;
}

/*****************************************************************************/

int TLPSendIncomingOutgoingAPDUBinary(TYPE_HANDLE fd,
				      unsigned char CLA,
				      unsigned char INS,
				      unsigned char P1,
				      unsigned char P2,
				      unsigned char L_in,
				      const unsigned char *data_in,
				      unsigned char L_out,
				      unsigned char *data_out,
				      unsigned char *ad_SW1,
				      unsigned char *ad_SW2)
{
	int resp_length;
	char buffer[TLP_DATA_BUF_SIZE];

	resp_length = TLPSendIncomingAPDUBinary(fd, CLA, INS, P1, P2,
					    L_in, data_in, ad_SW1, ad_SW2);
	
	if (resp_length == -1)
	{
	return -1;
	}
	

	/* Check value of SW 61 xx or not */
	if ( (*ad_SW1 != TLP_GTR_REQ_SW1))
	{
	rerrno = RD_CHAINING_ERROR;
	return -1;
	}
	
	/* Buffer is large enough to old max length : 256 bytes */
	resp_length = TLPSendOutgoingAPDUBinary(fd, 0x00, 0xC0, 0x00, 0x00,
					    *ad_SW2, buffer, ad_SW1,
					    ad_SW2);
	if (resp_length > L_out)
	{
	/* return a warning */
	rerrno = RD_RESP_TRUNCATED;
	resp_length = L_out;
	}
	
	memcpy(data_out, buffer, resp_length);

	return resp_length;

}


/*****************************************************************************/



int TLPSendIncomingAPDUAscii(TYPE_HANDLE fd, const char* APDU_Data,
			     char* ad_response, int resp_length)
{
	int cks_val;
	int data_length;
	char numb_buf[2];
	int i;
	int sw_length;
	char buffer[TLP_SW_BUF_SIZE];
	unsigned char rd_EOT;


	rd_EOT = TLP_EOT;
	/* 2 chars per Byte */
	data_length = strlen(APDU_Data) >> 1;
	data_length++; /* CMD byte */

	/* Checksum value */
	cks_val = TLP_ACK_BYTEh ^ data_length ^ TLP_IN_CMDh;
	for (i=0; i<(data_length-1); i++)
	{
	cks_val ^= c2bin(APDU_Data+2*i,2);
	}

	if (WRITE_COM(fd,TLP_ACK_BYTE, sizeof(TLP_ACK_BYTE))
	!= sizeof(TLP_ACK_BYTE))
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}

	bin2c(data_length, numb_buf, 2);
	if (WRITE_COM(fd,numb_buf, 2) != 2)
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}

	if (WRITE_COM(fd,TLP_IN_CMD, sizeof(TLP_IN_CMD))
	!= sizeof(TLP_IN_CMD))
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}

	data_length--;
	data_length <<= 1;
	if (WRITE_COM(fd,APDU_Data, data_length)
	!= data_length )
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}

	bin2c(cks_val, numb_buf, 2);
	if (WRITE_COM(fd,numb_buf, 2) != 2)
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}

	/* Send EOT to mark the end of the message */
	if (WRITE_COM(fd,&rd_EOT, 1) != 1)
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}


	/* Get card response */
	sw_length = TLPGetReaderResponse(fd, buffer, sizeof(buffer));
	if (sw_length == -1)
	{
	return -1;
	}

	/* Count in chars  */
	sw_length <<= 1;
	/* Add count for '\0' */
	sw_length++;

	resp_length = (sw_length<resp_length)?sw_length:resp_length;
	if (resp_length > 0)
	{
	memcpy(ad_response,buffer+TLP_DATA_OFF,resp_length-1);
	ad_response[resp_length-1] = '\0';
	}
	return resp_length;
}





/*****************************************************************************/



int TLPSendOutgoingAPDUAscii(TYPE_HANDLE fd, const char* APDU_Data,
			     char* ad_response, int resp_length)
{
	int cks_val;
	int data_length;
	char numb_buf[2];
	int i;
	int sw_length;
	char buffer[TLP_DATA_BUF_SIZE];
	unsigned char rd_EOT;


	rd_EOT = TLP_EOT;
	/* 2 chars per Byte */
	data_length = 5;  /* APDU bytes */
	data_length++;    /* CMD byte   */

	/* Checksum value */
	cks_val = TLP_ACK_BYTEh ^ data_length ^ TLP_OUT_CMDh;
	for (i=0; i<(data_length-1); i++)
	{
	cks_val ^= c2bin(APDU_Data+2*i,2);
	}

	if (WRITE_COM(fd,TLP_ACK_BYTE, sizeof(TLP_ACK_BYTE))
	!= sizeof(TLP_ACK_BYTE))
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}

	bin2c(data_length, numb_buf, 2);
	if (WRITE_COM(fd,numb_buf, 2) != 2)
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}

	if (WRITE_COM(fd,TLP_OUT_CMD, sizeof(TLP_OUT_CMD))
	!= sizeof(TLP_OUT_CMD))
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}

	if (WRITE_COM(fd,APDU_Data, 10)
	!= 10 )
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}


	bin2c(cks_val, numb_buf, 2);
	if (WRITE_COM(fd,numb_buf, 2) != 2)
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}

	/* Send EOT to mark the end of the message */
	if (WRITE_COM(fd,&rd_EOT, 1) != 1)
	{
	rerrno = RD_WRITE_FAILED;
	return -1;
	}


	/* Get card response */
	sw_length = TLPGetReaderResponse(fd, buffer, sizeof(buffer));
	if (sw_length == -1)
	{
	return -1;
	}

	/* Count in chars  */
	sw_length <<= 1;
	/* Add count for '\0' */
	sw_length++;

	resp_length = (sw_length<resp_length)?sw_length:resp_length;
	if (resp_length > 0)
	{
	memcpy(ad_response,buffer+TLP_DATA_OFF,resp_length-1);
	ad_response[resp_length-1] = '\0';
	}
	return resp_length;
}


/*****************************************************************************/



int TLPSendIncomingOutgoingAPDUAscii(TYPE_HANDLE fd, const char* APDU_Data,
				     char* ad_response, int resp_length)
{
	char buffer[TLP_DATA_BUF_SIZE];
	char buffer_cmd_gtr[11];
	int length;

	length = TLPSendIncomingAPDUAscii(fd, APDU_Data, buffer,
				      sizeof(buffer));
	if (length == -1)
	{
	return -1;
	}
	
	if ((buffer[length-5] != TLP_GTR_REQ_SW1_C1)
	||
	(buffer[length-4] != TLP_GTR_REQ_SW1_C2))
	{
	rerrno = RD_CHAINING_ERROR;
	return -1;
	}

	/* Append length to GTR command string */
	sprintf(buffer_cmd_gtr,"00C00000%s",buffer+length-3);

	length = TLPSendOutgoingAPDUAscii(fd, buffer_cmd_gtr,
				      buffer, sizeof(buffer));

	if (length == -1)
	{
	return -1;
	}

	if (resp_length < length)
	{
	rerrno = RD_RESP_TRUNCATED;
	length = resp_length;
	}
	
	if (length > 1)
	{
	memcpy(ad_response,buffer,length-1);
	}
	if (length > 0)
	{
  	ad_response[length-1] = '\0';
	}

	
	return length;

}
