
/*
 * shell utility to send commands to a card
 * JL Giraud, 10/99
 * L Rousseau, 09/2000
 *
 * $Id: scsend.c,v 1.6 2001/01/16 17:13:09 rousseau Exp $
 *
 */


#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>

#include "reader.h"

#define USAGE \
"Usage: scsend [-h][-v][-e][-t TYPE][-d DEVICE] [COMMAND] \n\
\n\
        -h:   displays this message\n\
        -d:   device connected to the reader (default /dev/screader)\n\
        -v:   output will also include the command sent to the card\n\
        -e:   include error interpreation accroding to ISO 7816-4\n\
        -t:   reader type (T for TLP, R for RF TLP, default is TLP )\n\
        COMMAND to send to the card (should be quoted) \n\
        The first character should be one of:\n\
                   'R': send a reset\n\
                   'I': INCOMING COMMAND (data sent to the card)\n\
                   'O': OUTGOING COMMAND (data received from the card\n\
                   'P': POWER DOWN\n\
                         ex: \"I 00 A4 00 00 02 3F 00\" (select MF)\n\
\n\
Send a ISO command to a card. If no command is given as parameter,\n\
commands are read from STDIN (one per line)\n"

#define BUF_SIZE 600
char default_device[] = "/dev/screader";

ReaderDesc rd;
char* _progname;


/*****************************************************************************
 *
 ****************************************************************************/
void send_error(char* info_string);
void tr_send_error(char* info_string);
void at_exit();
int send_command(char cmd_type, char *cmd_buffer, 
		 char* resp_buffer, int resp_length);
int format_buffer(const char* buffer_in, char* buffer_out);

/*****************************************************************************
 *
 ****************************************************************************/
int main(int argc, char *argv[])
{
    FILE *file_in, *file_out;
    char buffer[BUF_SIZE];
    char form_cmd[BUF_SIZE];
    char answer[BUF_SIZE];
    int res;
    char c;
    int verbose, device_opened, print_error;
    int length;
    char *command;
    int arg_count;
    rd_type reader_type;
    char *device;

    _progname = argv[0]; 

    file_in  = stdin;   /* Default value: command as a parameter */
    file_out = stdout;  /* Default value */
    device = default_device;  /* Default device should be used */
    verbose = 0;        /* By default, verbose mode is not set */
	print_error = 0;	/* do not print ISO 7816-4 errors */
    arg_count = 1;      /* argv[0] = program name */
    reader_type = TLP;
    device_opened = 0;

 
    /* Command line argument analysis */
    opterr = 0;
    while ((res = getopt(argc, argv, "t:d:hve")) != EOF) 
    { 
		c = (unsigned char ) res;
		switch(c) 
		{
		case 'd':

			arg_count++;
			arg_count++;
			/* Non default device requested */
			device = optarg;
			break;

		case 't':
			arg_count++;
			arg_count++;
			switch (tolower(argv[optind-1][0]))
			{
				case 't':
					reader_type = TLP;
					break;

				case 'r':
					reader_type = RFTLP;
					break;

				default:
					fprintf(stderr,"%s: %c: reader type unknown\n", 
						_progname, argv[optind-1][0]);
					exit(1);
					break;
			}
			break;

		case 'v':
			verbose = 1;
			arg_count++;
			break;
			
		case 'h':
			fprintf(stderr,USAGE);
			exit(0); 
			break;

		case 'e':
			print_error = 1;
			arg_count++;
			break;

		default:
			fprintf(stderr,USAGE);
			exit(1); 
		}
    }
  
    /* Check if defaut device should be used */
    if (InitReader(&rd, reader_type, device) == -1)
    {
		send_error(device);
		exit(1);
    }


    /* Register exit function */
    /* This is just to have a clean way to exit */
    if (atexit(at_exit))
    {
		at_exit();
		send_error("Could not register exit function");
		exit(1);
    }
  
    if (argc > arg_count)
    {
		/* Command was given as parameter */
		/* Check the size (-1) for '\0'   */
		command = argv[optind];
		puts(command);
		if (strlen(command) > (BUF_SIZE-1))
		{
			fprintf(stderr,"%s: command too long: %s\n",_progname,command);
			exit(1);
		}

		if ((length = format_buffer(command, form_cmd)) == -1)
		{
			return -1;
		}

		if (length %2)
		{
			fprintf(stderr,"%s: odd length for: %s\n",_progname,command);
			exit(1);
		}
		  
		/* BUF_SIZE-1 for '\0' */
		if (send_command(command[0], form_cmd, answer, BUF_SIZE-1) == -1)
		{
			exit(1);
		}
		
		/* ouput result */
		if (verbose)
		{
			fprintf(file_out,"\nIN: %s\nOUT: ", command);
		}

		fprintf(file_out, "%s\n", answer);

		/* ISO 7816-4 error */
		switch (tolower(command[0]))
		{
			case 'i':
			case 'o':
				{
					int sw1, sw2;

					sscanf(answer + strlen(answer) - 4, "%02X%02X", &sw1, &sw2);
					fprintf(file_out, "%s\n", ISO7816_ErrorStr(sw1, sw2));
				}
		}

		exit(0);
    }

    /* read commands from file_in */
    while (fgets(buffer, BUF_SIZE, file_in) != NULL)
    {
		if (buffer[0] == '#')
		{
			/* This line was a comment */
			continue;
		}
		if ((length = format_buffer(buffer, form_cmd)) == -1)
		{
			return -1;
		}
		if (length %2)
		{
			fprintf(stderr,"%s: odd length for: %s\n",_progname,buffer);
			exit(1);
		}
		  
		/* BUF_SIZE-1 for '\0' */
		if (send_command(buffer[0], form_cmd, answer, BUF_SIZE-1) == -1)
		{
			exit(1);
		}
		/* ouput result */
		if (verbose)
		{
			fprintf(file_out,"\nIN: %sOUT: ", buffer);
		}
		fprintf(file_out, "%s\n", answer);

		/* ISO 7816-4 error */
		switch (tolower(buffer[0]))
		{
			case 'i':
			case 'o':
				{
					int sw1, sw2;

					sscanf(answer + strlen(answer) - 4, "%02X%02X", &sw1, &sw2);
					fprintf(file_out, "%s\n", ISO7816_ErrorStr(sw1, sw2));
				}
		}
    }

    exit(0);
  
} /* main */


/*****************************************************************************
 *
 ****************************************************************************/
int send_command(char cmd_type, char *cmd_buffer, 
		 char* resp_buffer, int resp_length)
{
    int length;

    switch(tolower(cmd_type))
    {
		case 'r':
			if ((length = ResetCardAscii(&rd,resp_buffer,resp_length)) == -1)
			{
				tr_send_error("reset failed");
				return -1;
			}
			return length;
			break;

		case 'i':
			if ( (length = SendIncomingAPDUAscii(&rd, cmd_buffer,resp_buffer, 
								 resp_length)==-1))
			{
				tr_send_error("incoming command failed");
				return -1;
			}
			return length;
			break;

		case 'o':
			if ( (length = SendOutgoingAPDUAscii(&rd, cmd_buffer,resp_buffer, 
								 resp_length)==-1))
			{
				tr_send_error("outgoing command failed");
				return -1;
			}
			return length;
			break;

		case 'p':
			if (PowerDown(&rd) == -1)
			{
				tr_send_error("power down failed");
				return -1;
			}
			if (resp_length)
				resp_buffer[0] = '\0';
			return 0;

		default:
			fprintf(stderr,"%s, unknown command %c\n",_progname, cmd_type);
			return -1;
			break;
    }

    fprintf(stderr,"%s, how the Hell did you get here??\n",_progname);

    return -1;
} /* send_command */


/*****************************************************************************
 *
 ****************************************************************************/
int format_buffer(const char* buffer_in, char* buffer_out)
{
    int length,i, index;

    /* Formats the buffer (removing spaces,...)    */
    /* buffer_in should be '\0' terminated         */
    /* returns length of buffer_out without '\0'   */
    length = strlen(buffer_in);
    index  = 0; 
  
    for (i=0; i<length; i++)
    {
	if (isxdigit(buffer_in[i]))
	{
	    buffer_out[index++] = toupper(buffer_in[i]);
	}
    }
    buffer_out[index] = '\0';
    return index;
} /* format_buffer */


/*****************************************************************************
 *
 ****************************************************************************/
void tr_send_error(char* info_string)
{
    fprintf(stderr,"%s: %s: %s\n", _progname, info_string,
		GetRerrStr(GetRerrno()));
} /* tr_send_error */


/*****************************************************************************
 *
 ****************************************************************************/
void send_error(char* info_string)
{
    fprintf(stderr,"%s: %s: %s\n", _progname, info_string, strerror(errno));
} /* send_error */


/*****************************************************************************
 *
 ****************************************************************************/
void at_exit(void)
{
    CloseReader(&rd);
} /* at_exit */
